use std::collections::{HashMap, HashSet};
use std::iter::zip;

use hercules_ir::ir::*;

use crate::*;

/*
 * Split multi-dimensional fork-joins into separate one-dimensional fork-joins.
 * Useful for code generation.
 */
pub fn fork_split(
    editor: &mut FunctionEditor,
    fork_join_map: &HashMap<NodeID, NodeID>,
    reduce_cycles: &HashMap<NodeID, HashSet<NodeID>>,
) {
    // A single multi-dimensional fork becomes multiple forks, a join becomes
    // multiple joins, a thread ID becomes a thread ID on the correct
    // fork, and a reduce becomes multiple reduces to shuffle the reduction
    // value through the fork-join nest.
    for (fork, join) in fork_join_map {
        let nodes = &editor.func().nodes;
        let (fork_control, factors) = nodes[fork.idx()].try_fork().unwrap();
        if factors.len() < 2 {
            continue;
        }
        let factors: Box<[DynamicConstantID]> = factors.into();
        let join_control = nodes[join.idx()].try_join().unwrap();
        let tids: Vec<_> = editor
            .get_users(*fork)
            .filter(|id| nodes[id.idx()].is_thread_id())
            .collect();
        let reduces: Vec<_> = editor
            .get_users(*join)
            .filter(|id| nodes[id.idx()].is_reduce())
            .collect();

        let data_in_reduce_cycle: HashSet<(NodeID, NodeID)> = reduces
            .iter()
            .map(|reduce| editor.get_users(*reduce).map(move |user| (user, *reduce)))
            .flatten()
            .filter(|(user, reduce)| reduce_cycles[&reduce].contains(&user))
            .collect();

        editor.edit(|mut edit| {
            // Create the forks and a thread ID per fork.
            let mut acc_fork = fork_control;
            let mut new_tids = vec![];
            for factor in factors {
                acc_fork = edit.add_node(Node::Fork {
                    control: acc_fork,
                    factors: Box::new([factor]),
                });
                edit.sub_edit(*fork, acc_fork);
                new_tids.push(edit.add_node(Node::ThreadID {
                    control: acc_fork,
                    dimension: 0,
                }));
            }

            // Create the joins.
            let mut acc_join = if join_control == *fork {
                acc_fork
            } else {
                join_control
            };
            let mut joins = vec![];
            for _ in new_tids.iter() {
                acc_join = edit.add_node(Node::Join { control: acc_join });
                edit.sub_edit(*join, acc_join);
                joins.push(acc_join);
            }

            // Create the reduces.
            let mut new_reduces = vec![];
            for reduce in reduces.iter() {
                let (_, init, reduct) = edit.get_node(*reduce).try_reduce().unwrap();
                let num_nodes = edit.num_node_ids();
                let mut inner_reduce = NodeID::new(0);
                let mut outer_reduce = NodeID::new(0);
                for (join_idx, join) in joins.iter().enumerate() {
                    let init = if join_idx == joins.len() - 1 {
                        init
                    } else {
                        NodeID::new(num_nodes + join_idx + 1)
                    };
                    let reduct = if join_idx == 0 {
                        reduct
                    } else {
                        NodeID::new(num_nodes + join_idx - 1)
                    };
                    let new_reduce = edit.add_node(Node::Reduce {
                        control: *join,
                        init,
                        reduct,
                    });
                    assert_eq!(new_reduce, NodeID::new(num_nodes + join_idx));
                    edit.sub_edit(*reduce, new_reduce);
                    if join_idx == 0 {
                        inner_reduce = new_reduce;
                    }
                    if join_idx == joins.len() - 1 {
                        outer_reduce = new_reduce;
                    }
                }
                new_reduces.push((inner_reduce, outer_reduce));
            }

            // Replace everything.
            edit = edit.replace_all_uses(*fork, acc_fork)?;
            edit = edit.replace_all_uses(*join, acc_join)?;
            for tid in tids.iter() {
                let dim = edit.get_node(*tid).try_thread_id().unwrap().1;
                edit.sub_edit(*tid, new_tids[dim]);
                edit = edit.replace_all_uses(*tid, new_tids[dim])?;
            }
            for (reduce, (inner_reduce, outer_reduce)) in zip(reduces.iter(), new_reduces) {
                edit = edit.replace_all_uses_where(*reduce, inner_reduce, |id| {
                    data_in_reduce_cycle.contains(&(*id, *reduce))
                })?;
                edit = edit.replace_all_uses_where(*reduce, outer_reduce, |id| {
                    !data_in_reduce_cycle.contains(&(*id, *reduce))
                })?;
            }

            // Delete all the old stuff.
            edit = edit.delete_node(*fork)?;
            edit = edit.delete_node(*join)?;
            for tid in tids {
                edit = edit.delete_node(tid)?;
            }
            for reduce in reduces {
                edit = edit.delete_node(reduce)?;
            }

            Ok(edit)
        });
    }
}
